import os

from opentelemetry.instrumentation.distro import BaseDistro
from opentelemetry.environment_variables import OTEL_TRACES_EXPORTER

from .uptrace import configure_opentelemetry

# pylint: disable=too-few-public-methods
class UptraceDistro(BaseDistro):
    def _configure(self, **kwargs):
        configure_opentelemetry()



import logging
import os
from typing import Optional

from opentelemetry import trace

from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace.export import BatchSpanProcessor
from opentelemetry.sdk.resources import Attributes, Resource

from .client import Client
from .dsn import parse_dsn, DSN
from .spanexp import UptraceSpanExporter

logger = logging.getLogger(__name__)

_CLIENT = None
_FALLBACK_CLIENT = Client(parse_dsn("https://<token>@api.uptrace.dev/<project_id>"))

# pylint: disable=too-many-arguments
def configure_opentelemetry(
    dsn="",
    service_name: Optional[str] = "",
    service_version: Optional[str] = "",
    resource_attributes: Optional[Attributes] = None,
    resource: Optional[Resource] = None,
):
    """
    configureOpentelemetry configures OpenTelemetry to export data to Uptrace.
    By default it:
       - creates tracer provider;
       - registers Uptrace span exporter.
    """

    global _CLIENT  # pylint: disable=global-statement

    if os.getenv("UPTRACE_DISABLED") == "True":
        return

    if _CLIENT is not None:
        logger.warning("Uptrace is already configured")

    if not dsn:
        dsn = os.getenv("UPTRACE_DSN", "")

    try:
        dsn = parse_dsn(dsn)
    except ValueError as exc:
        # pylint:disable=logging-not-lazy
        logger.warning("Uptrace is disabled: %s", exc)
        return

    _CLIENT = Client(dsn=dsn)
    _configure_tracing(
        dsn,
        service_name=service_name,
        service_version=service_version,
        resource_attributes=resource_attributes,
        resource=resource,
    )


def _configure_tracing(
    dsn: DSN,
    service_name: Optional[str] = "",
    service_version: Optional[str] = "",
    resource_attributes: Optional[Attributes] = None,
    resource: Optional[Resource] = None,
):
    if trace._TRACER_PROVIDER is None:
        resource = _build_resource(
            resource, resource_attributes, service_name, service_version
        )
        provider = TracerProvider(resource=resource)
        trace.set_tracer_provider(provider)

    exporter = UptraceSpanExporter(dsn)
    bsp = BatchSpanProcessor(
        exporter,
        max_queue_size=1000,
        max_export_batch_size=1000,
        schedule_delay_millis=5000,
    )
    trace.get_tracer_provider().add_span_processor(bsp)


def trace_url(span: Optional[trace.Span] = None) -> str:
    """Returns the trace URL for the span."""

    if _CLIENT is not None:
        return _CLIENT.trace_url(span)

    return _FALLBACK_CLIENT.trace_url(span)


def report_exception(exc: Exception) -> None:
    if _CLIENT is not None:
        _CLIENT.report_exception(exc)


def _build_resource(
    resource: Resource,
    resource_attributes: Attributes,
    service_name: str,
    service_version: str,
) -> Resource:
    attrs = {}

    if resource_attributes:
        attrs.update(resource_attributes)
    if service_name:
        attrs["service.name"] = service_name
    if service_version:
        attrs["service.version"] = service_version

    if resource is None:
        return Resource.create(attrs)

    if len(attrs) == 0:
        return resource

    return resource.merge(Resource.create(attrs))



"""Uptrace span exporter for OpenTelemetry"""

import logging
import typing
from types import MappingProxyType

import msgpack
import requests
import zstd
from opentelemetry.sdk import trace as trace_sdk
from opentelemetry.sdk.trace import export as sdk
from opentelemetry.sdk.util import BoundedDict
from opentelemetry.trace import Link, SpanKind
from opentelemetry.trace.status import StatusCode

from ..dsn import DSN

logger = logging.getLogger(__name__)


class UptraceSpanExporter(
    sdk.SpanExporter
):  # pylint:disable=too-many-instance-attributes
    """Uptrace span exporter for OpenTelemetry."""

    def __init__(self, dsn: DSN):
        self._dsn = dsn
        self._closed = False

        self._endpoint = (
            f"{dsn.scheme}://{dsn.host}/api/v1/tracing/{dsn.project_id}/spans"
        )
        self._headers = {
            "Authorization": "Bearer " + dsn.token,
            "Content-Type": "application/msgpack",
            "Content-Encoding": "zstd",
        }

    def export(self, spans: typing.Sequence[sdk.Span]) -> sdk.SpanExportResult:
        if self._closed:
            return sdk.SpanExportResult.SUCCESS

        spans_out = []

        for span in spans:
            out = _out_span(span)
            spans_out.append(out)

        self._send(spans_out)

        return sdk.SpanExportResult.SUCCESS

    def shutdown(self) -> None:
        if self._closed:
            return
        self._closed = True

    def _send(self, spans):
        payload = msgpack.packb({"spans": spans})
        payload = zstd.compress(payload)

        resp = requests.post(self._endpoint, data=payload, headers=self._headers)
        if resp.status_code < 200 or resp.status_code >= 300:
            logger.error("uptrace: status=%d %s", resp.status_code, resp.text)


def _out_span(span: sdk.Span):
    out = {
        "id": span.context.span_id,
        "traceId": _trace_id_bytes(span.context.trace_id),
        "name": span.name,
        "kind": _kind(span.kind),
        "startTime": span.start_time,
        "endTime": span.end_time,
    }

    if span.parent is not None:
        out["parentId"] = span.parent.span_id

    out["statusCode"] = _status(span.status.status_code)
    if span.status.description:
        out["statusMessage"] = span.status.description

    out["tracerName"] = span.instrumentation_info.name
    if span.instrumentation_info.version:
        out["tracerVersion"] = span.instrumentation_info.version

    if span.attributes:
        out["attrs"] = _attrs(span.attributes)

    if span.events:
        out["events"] = _events(span.events)

    if span.links:
        out["links"] = _links(span.links)

    if span.resource:
        out["resource"] = _attrs(span.resource.attributes)

    return out


def _events(events: typing.Sequence[trace_sdk.Event]):
    out = []
    for event in events:
        out.append(
            {
                "name": event.name,
                "attrs": _attrs(event.attributes),
                "time": event.timestamp,
            }
        )
    return out


def _links(links: typing.Sequence[Link]):
    out = []
    for link in links:
        out.append(
            {
                "traceId": _trace_id_bytes(link.context.trace_id),
                "spanId": link.context.span_id,
                "attrs": _attrs(link.attributes),
            }
        )
    return out


def _kind(kind: SpanKind) -> str:
    if kind == SpanKind.SERVER:
        return "server"
    if kind == SpanKind.CLIENT:
        return "client"
    if kind == SpanKind.PRODUCER:
        return "producer"
    if kind == SpanKind.CONSUMER:
        return "consumer"
    return "internal"


def _status(status: StatusCode) -> str:
    if status == StatusCode.ERROR:
        return "error"
    if status == StatusCode.OK:
        return "ok"
    return "unset"


def _attrs(attrs):
    if isinstance(attrs, BoundedDict):
        return attrs._dict  # pylint: disable=protected-access
    if isinstance(attrs, MappingProxyType):
        return dict(attrs)
    return attrs


def _trace_id_bytes(trace_id: int):
    return trace_id.to_bytes(16, byteorder="big")



from collections import namedtuple
from urllib.parse import urlparse

from .util import remove_prefix

DSN = namedtuple("DSN", ["scheme", "host", "project_id", "token"])


def parse_dsn(dsn: str) -> DSN:
    if dsn == "":
        raise ValueError("either dsn option or UPTRACE_DSN is required")

    o = urlparse(dsn)
    if not o.scheme:
        raise ValueError(f"can't parse DSN={dsn}")

    host = o.hostname
    if not host:
        raise ValueError(f"DSN={dsn} does not have a host")

    if o.port:
        host += f":{o.port}"

    project_id = remove_prefix(o.path, "/")
    if not project_id:
        raise ValueError(f"DSN={dsn} does not have a project id")

    token = o.username
    if not token:
        raise ValueError(f"DSN={dsn} does not have a token")

    return DSN(
        scheme=o.scheme,
        host=host,
        project_id=project_id,
        token=token,
    )



"""Uptrace exporter for OpenTelemetry"""

from .uptrace import configure_opentelemetry, report_exception, trace_url
from .version import __version__

__all__ = ["configure_opentelemetry", "trace_url", "report_exception", "__version__"]



from unittest.mock import patch

import pytest
from opentelemetry import trace as trace
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.trace.status import StatusCode

import uptrace
from uptrace.spanexp import UptraceSpanExporter


def setup_function():
    trace._TRACER_PROVIDER = TracerProvider()


def teardown_function():
    trace._TRACER_PROVIDER = None


def test_span_processor_no_dsn(caplog):
    uptrace.configure_opentelemetry()
    assert "either dsn option or UPTRACE_DSN is required" in caplog.text


def test_span_processor_invalid_dsn(caplog):
    uptrace.configure_opentelemetry(dsn="invalid")
    assert "can't parse DSN=invalid" in caplog.text


def test_trace_url():
    uptrace.configure_opentelemetry(dsn="https://token@api.uptrace.dev/123")
    tracer = trace.get_tracer("tracer_name")
    span = tracer.start_span("main span")

    url = uptrace.trace_url(span)
    assert url.startswith("https://uptrace.dev/search/123?q=")


@patch.object(UptraceSpanExporter, "_send")
def test_send(send):
    uptrace.configure_opentelemetry(
        dsn="https://<token>@api.uptrace.dev/<project_id>",
        service_name="myservice",
    )

    uptrace.report_exception(ValueError("hello"))
    trace.get_tracer_provider().shutdown()

    send.assert_called()
    spans = send.call_args[0][0]
    assert len(spans) == 1, traces
    span = spans[0]

    assert type(span["startTime"]) is int
    assert type(span["endTime"]) is int

    assert span["kind"] == "internal"
    assert span["statusCode"] == "unset"

    assert span["tracerName"] == "uptrace-python"

    assert span["resource"] == {
        "service.name": "unknown_service",
        "telemetry.sdk.language": "python",
        "telemetry.sdk.name": "opentelemetry",
        "telemetry.sdk.version": "1.1.0",
    }

    events = span["events"]
    assert len(events) == 1

    event = events[0]
    attrs = event["attrs"]
    assert attrs["exception.type"] == "ValueError"
    assert attrs["exception.message"] == "hello"
    assert attrs["exception.stacktrace"] == "NoneType: None\n"
